// Copyright 2022 Tilt Five, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "HMD/TiltFiveHMD.h"

#include "ClearQuad.h"
#include "DefaultSpectatorScreenController.h"
#include "PipelineStateCache.h"
#include "ScreenRendering.h"
#if UE_VERSION_NEWER_THAN(4, 21, 2)
#include "CommonRenderResources.h"
#endif

#ifndef UE_ARRAY_COUNT
#define UE_ARRAY_COUNT ARRAY_COUNT
#endif

#define ENGINE_SPECIFIC_HEADER_INNER(Prefix, Major, Minor, Suffix) \
	PREPROCESSOR_TO_STRING(PREPROCESSOR_JOIN(                      \
		PREPROCESSOR_JOIN(PREPROCESSOR_JOIN(PREPROCESSOR_JOIN(Prefix, _), Major), _), PREPROCESSOR_JOIN(Minor, Suffix)))
// #include ENGINE_SPECIFIC_HEADER(Example) creates an include command appending the current engine,
// e.g. for 4.20 it creates #include "Example_4_20.h"
#define ENGINE_SPECIFIC_HEADER(Prefix) ENGINE_SPECIFIC_HEADER_INNER(Prefix, ENGINE_MAJOR_VERSION, ENGINE_MINOR_VERSION, .h)

// clang-format off
#include ENGINE_SPECIFIC_HEADER(Engine/TiltFiveHMD)
// clang-format on

#include "Engine/Canvas.h"
#include "Engine/Engine.h"
#include "GameFramework/WorldSettings.h"
#include "Modules/ModuleManager.h"
#include "PipelineStateCache.h"
#include "RendererInterface.h"
#include "XRThreadUtils.h"

#include <chrono>
#include <thread>

static const int32 GMaxNumTiltFiveGlasses = 4;

const FQuat rotToUGLS_GLS = FQuat(0, 0.7071068, 0, -0.7071068);
const FQuat rotToGLS_UGLS = FQuat(0, 0.7071068, 0, 0.7071068);

FQuat FTiltFiveHMD::ConvertFromHardware(const FQuat& rotToGLS_GBD)
{
	// Swap X and Y Axes to transform axis of rotation to Tilt Five GBD from Unreal World Space, and negate W component to swap Tilt
	// Five's RHS to Unreal's LHS
	FQuat Result = FQuat(rotToGLS_GBD.Y, rotToGLS_GBD.X, rotToGLS_GBD.Z, -rotToGLS_GBD.W);
	// Rotate to Unreal camera orientation from Tilt Five camera orientation.
	FQuat Result2 = rotToUGLS_GLS * Result;
	// Negate W component because Unreal requires Rotation from Camera space to world space, whereas Tilt Five provides a rotation
	// that takes point in world space to camera space.
	return FQuat(Result2.X, Result2.Y, Result2.Z, -Result2.W);
}

FQuat FTiltFiveHMD::ConvertToHardware(const FQuat& rotToGLS_GBD)
{
	// Negate W component because Unreal provides Rotation from Camera space to world space, whereas Tilt Five expect a rotation
	// that takes point in world space to camera space.
	FQuat Result = FQuat(rotToGLS_GBD.X, rotToGLS_GBD.Y, rotToGLS_GBD.Z, -rotToGLS_GBD.W);
	// Rotate from Unreal camera orientation to Tilt Five camera orientation.
	FQuat Result2 = rotToGLS_UGLS * Result;
	// Swap X and Y Axes to transform axis of rotation from Tilt Five GBD to Unreal World Space, and negate W component to swap
	// Unreals LHS to Tilt Five's RHS
	return FQuat(Result2.Y, Result2.X, Result2.Z, -Result2.W);
}

FVector FTiltFiveHMD::ConvertFromHardware(const FVector& Location, float WorldToMetersScale)
{
	return FVector(Location.Y, Location.X, Location.Z) * WorldToMetersScale;
}

FVector FTiltFiveHMD::ConvertToHardware(const FVector& Location, float WorldToMetersScale)
{
	return FVector(Location.Y, Location.X, Location.Z) / WorldToMetersScale;
}

FTiltFiveHMD::FTiltFiveHMD(const FAutoRegister& Register)
	:
#if UE_VERSION_NEWER_THAN(4, 21, 2)
	FHeadMountedDisplayBase(nullptr)	// TODO(marvin@lab132.com) Maybe implement support for AR Tracking (T5 Cards?)
#else
	FHeadMountedDisplayBase()
#endif
	, FSceneViewExtensionBase(Register)
{
	for (int32 EyeIndex = 0; EyeIndex < NumEyeRenderTargets; ++EyeIndex)
	{
		FMemory::Memzero(EyeInfos[EyeIndex]);
	}
}

FTiltFiveHMD::~FTiltFiveHMD()
{
	for (int32 EyeIndex = 0; EyeIndex < NumEyeRenderTargets; ++EyeIndex)
	{
		EyeInfos[EyeIndex].BufferedRTRHI = nullptr;
		EyeInfos[EyeIndex].BufferedSRVRHI = nullptr;
	}
}

#if UE_VERSION_NEWER_THAN(4, 25, 4)
int32 FTiltFiveHMD::GetXRSystemFlags() const
{
	return EXRSystemFlags::IsHeadMounted;
}
#endif

bool FTiltFiveHMD::EnumerateTrackedDevices(TArray<int32>& OutDevices, EXRTrackedDeviceType Type /*= EXRTrackedDeviceType::Any*/)
{
	if (Type == EXRTrackedDeviceType::HeadMountedDisplay || Type == EXRTrackedDeviceType::Any)
	{
		OutDevices.Add(HMDDeviceId);
		return true;
	}
	return false;
}

void FTiltFiveHMD::UpdateCachedGlassesPose_RenderThread()
{
	check(IsInRenderingThread());

	if (!RenderThreadWorldState.IsValid())
	{
		CachedGlassesPoseIsValid_RenderThread = false;
		return;
	}

	FScopeLock ScopeLock(&ExclusiveGroup1CriticalSection);

	FT5GlassesPose Pose;

	const FT5Result Result = t5GetGlassesPose(CurrentExclusiveGlasses, kT5_GlassesPoseUsage_GlassesPresentation, &Pose);

	if (Result == T5_SUCCESS)
	{
		CachedGlassesPoseIsValid_RenderThread = true;
		CachedGlassesOrientation_RenderThread = ConvertRotationFromHardware(Pose.rotToGLS_GBD);
		CachedGlassesPosition_RenderThread =
			ConvertPositionFromHardware(Pose.posGLS_GBD, RenderThreadWorldState->WorldToMetersScale);
		CachedGameboardType_RenderThread = Pose.gameboardType;
	}
	else
	{
		CachedGlassesPoseIsValid_RenderThread = false;
	}
}

void FTiltFiveHMD::UpdateCachedGlassesPose_GameThread()
{
	check(IsInGameThread());

	if (!CurrentWorldState.IsValid())
	{
		CachedGlassesPoseIsValid_GameThread = false;
		return;
	}

	FScopeLock ScopeLock(&ExclusiveGroup1CriticalSection);

	FT5GlassesPose Pose;

	const FT5Result Result = t5GetGlassesPose(CurrentExclusiveGlasses, kT5_GlassesPoseUsage_GlassesPresentation, &Pose);

	if (Result == T5_SUCCESS)
	{
		CachedGlassesPoseIsValid_GameThread = true;
		CachedGlassesOrientation_GameThread = ConvertRotationFromHardware(Pose.rotToGLS_GBD);
		CachedGlassesPosition_GameThread = ConvertPositionFromHardware(Pose.posGLS_GBD, CurrentWorldState->WorldToMetersScale);
		CachedGameboardType_GameThread = Pose.gameboardType;
	}
	else
	{
		CachedGlassesPoseIsValid_GameThread = false;
	}
}

bool FTiltFiveHMD::GetCurrentPose(int32 DeviceId, FQuat& OutOrientation, FVector& OutPosition)
{
	if (DeviceId == ETiltFiveDeviceId::Glasses)
	{
		if (IsInRenderingThread() && CachedGlassesPoseIsValid_RenderThread)
		{
			OutOrientation = CachedGlassesOrientation_RenderThread;
			OutPosition = CachedGlassesPosition_RenderThread;
			return true;
		}
		else if (IsInGameThread() && CachedGlassesPoseIsValid_GameThread)
		{
			OutOrientation = CachedGlassesOrientation_GameThread;
			OutPosition = CachedGlassesPosition_GameThread;
			return true;
		}
	}

	OutOrientation = FQuat::Identity;
	OutPosition = FVector::ZeroVector;
	return false;
}

float FTiltFiveHMD::GetWorldToMetersScale() const
{
	FTiltFiveWorldStatePtr CurrentState;
	if (IsInRenderingThread())
	{
		CurrentState = RenderThreadWorldState;
	}
	else if (IsInGameThread())
	{
		CurrentState = CurrentWorldState;
	}
	else
	{
		// Fallback to default value
		return 100.0f;
	}

	if (!CurrentState.IsValid())
	{
		// Fallback to default value
		return 100.0f;
	}

	return CurrentState->WorldToMetersScale;
}

void FTiltFiveHMD::ResetOrientationAndPosition(float Yaw /*= 0.f*/)
{
	unimplemented();
}

const FName FTiltFiveHMD::TiltFiveSystemName(TEXT("TiltFiveHMD"));

bool FTiltFiveHMD::IsActiveThisFrame(class FViewport* InViewport) const
{
	return IsStereoEnabled();
}

void FTiltFiveHMD::Startup()
{
	SpectatorScreenController = MakeUnique<FDefaultSpectatorScreenController>(this);
	SpectatorScreenController->SetSpectatorScreenMode(ESpectatorScreenMode::SingleEye);

	static const FName RendererModuleName("Renderer");
	RendererModule = FModuleManager::GetModulePtr<IRendererModule>(RendererModuleName);

	CustomPresent = new FTiltFiveCustomPresent(*this);
}

void FTiltFiveHMD::RenderTexture_RenderThread(class FRHICommandListImmediate& RHICmdList,
#if UE_VERSION_OLDER_THAN(5, 1, 0)
		class FRHITexture2D* BackBuffer,
		class FRHITexture2D* SrcTexture,
#else
		class FRHITexture* BackBuffer,
		class FRHITexture* SrcTexture,
#endif
		FVector2D WindowSize) const
{
	for (int32 EyeIndex = 0; EyeIndex < NumEyeRenderTargets; ++EyeIndex)
	{
		const FIntRect DestEyeRect(
			0, 0, EyeInfos[EyeIndex].BufferedRTRHI->GetSizeX(), EyeInfos[EyeIndex].BufferedRTRHI->GetSizeY());
		CopyTexture_RenderThread(RHICmdList,
			SrcTexture,
			EyeInfos[EyeIndex].SourceEyeRect,
			EyeInfos[EyeIndex].BufferedRTRHI,
			DestEyeRect,
			false
#if UE_VERSION_NEWER_THAN(4, 20, 3)
			,
			true
#endif
		);

#if UE_VERSION_NEWER_THAN(4, 25, 4)
		RHICmdList.Transition(FRHITransitionInfo(EyeInfos[EyeIndex].BufferedSRVRHI, ERHIAccess::WritableMask, ERHIAccess::SRVMask));
#else
		FRHITexture* Texture = EyeInfos[EyeIndex].BufferedRTRHI;
		RHICmdList.TransitionResources(EResourceTransitionAccess::EReadable, &Texture, 1);
#endif
	}
	if (SpectatorScreenController)
	{
		SpectatorScreenController->RenderSpectatorScreen_RenderThread(RHICmdList, BackBuffer, SrcTexture, WindowSize);
	}
}

IStereoRenderTargetManager* FTiltFiveHMD::GetRenderTargetManager()
{
	return this;
}

FIntPoint FTiltFiveHMD::GetIdealRenderTargetSize() const
{
	return FIntPoint(2432, 768);
}

bool FTiltFiveHMD::GetHMDDistortionEnabled(EShadingPath ShadingPath) const
{
	// No (software) distortion necessary for tilt five
	return false;
}

#if UE_VERSION_NEWER_THAN(4, 25, 4)
bool FTiltFiveHMD::AllocateRenderTargetTexture(uint32 Index,
	uint32 SizeX,
	uint32 SizeY,
	uint8 Format,
	uint32 NumMips,
	ETextureCreateFlags Flags,
	ETextureCreateFlags TargetableTextureFlags,
	FTexture2DRHIRef& OutTargetableTexture,
	FTexture2DRHIRef& OutShaderResourceTexture,
	uint32 NumSamples /*= 1*/)
{
	FRHIResourceCreateInfo CreateInfo{TEXT("TiltFiveCreateInfo")};
	const int32 EyeSizeX = SizeX / 2;
	const int32 EyeSizeY = SizeY;
	WidthToHeight = (float) EyeSizeX / (float) EyeSizeY;

	for (int32 EyeIndex = 0; EyeIndex < NumEyeRenderTargets; ++EyeIndex)
	{
#if !UE_BUILD_SHIPPING
		EyeInfos[EyeIndex].DebugName = FString::Printf(TEXT("TiltFiveEyeRT%d"), EyeIndex);
		CreateInfo.DebugName = *EyeInfos[EyeIndex].DebugName;
#endif
		EyeInfos[EyeIndex].SourceEyeRect = FIntRect(EyeSizeX * EyeIndex, 0, EyeSizeX * EyeIndex + EyeSizeX, EyeSizeY);
		EyeInfos[EyeIndex].DestEyeRect = FIntRect(0, 0, EyeSizeX, EyeSizeY);
		RHICreateTargetableShaderResource2D(EyeSizeX,
			EyeSizeY,
			PF_R8G8B8A8,
			1,
			TexCreate_None,
			TexCreate_RenderTargetable,
			false,
			CreateInfo,
			EyeInfos[EyeIndex].BufferedRTRHI,
			EyeInfos[EyeIndex].BufferedSRVRHI);
	}

	// Create the native buffer for rendering using the default handling, so return false here
	return false;
}
#endif

void FTiltFiveHMD::SetTrackingOrigin(EHMDTrackingOrigin::Type NewOrigin)
{
	if (NewOrigin != EHMDTrackingOrigin::Floor)
	{
		UE_LOG(LogTiltFive, Error, TEXT("Tilt Five only supports Floor Level Tracking Origin"));
	}
}

#if UE_VERSION_NEWER_THAN(4, 22, 3)
EHMDTrackingOrigin::Type FTiltFiveHMD::GetTrackingOrigin() const
{
	return EHMDTrackingOrigin::Floor;
}
#endif

bool FTiltFiveHMD::OnStartGameFrame(FWorldContext& WorldContext)
{
	const UWorld* World = WorldContext.World();

	if (!World || !World->IsGameWorld())
	{
		// ignore all non-game worlds
		return false;
	}

	const float CurrentTime = static_cast<float>(FApp::GetCurrentTime());
	// If the current time for some reason gets reset, we also reset our timer,
	// otherwise we may wait a very long time here
	if (LastGlassesUpdate > CurrentTime)
	{
		LastGlassesUpdate = CurrentTime;
	}
	// Poll Glasses once per second
	if (LastGlassesUpdate + 1.0f < CurrentTime)
	{
		LastGlassesUpdate = CurrentTime;
		ConditionalInitializeGlasses();
	}

	CurrentWorldState.Reset();

	if (const AWorldSettings* WorldSettings = WorldContext.World()->GetWorldSettings())
	{
		FTiltFiveWorldState WorldState;
		WorldState.WorldToMetersScale = WorldSettings->WorldToMeters;
		CurrentWorldState = MakeShared<FTiltFiveWorldState, ESPMode::ThreadSafe>(WorldState);
	}

	UpdateCachedGlassesPose_GameThread();

	FQuat GlassesOrientation;
	FVector GlassesPosition;
	TOptional<FTransform> MaybeRelativeGlassesTransform;

	if (GetCurrentPose(HMDDeviceId, GlassesOrientation, GlassesPosition))
	{
		MaybeRelativeGlassesTransform = FTransform(GlassesOrientation, GlassesPosition);
	}

	FTiltFiveWorldStatePtr RenderThreadCopy = MakeShared<FTiltFiveWorldState, ESPMode::ThreadSafe>(*CurrentWorldState);

	ExecuteOnRenderThread_DoNotWait(
		[this, RenderThreadCopy, MaybeRelativeGlassesTransform](FRHICommandListImmediate& RHICmdList)
		{
			this->RenderThreadWorldState = RenderThreadCopy;
			this->MaybeRelativeGlassesTransform_RenderThread = MaybeRelativeGlassesTransform;
		});

	RefreshTrackingToWorldTransform(WorldContext);

	return FXRTrackingSystemBase::OnStartGameFrame(WorldContext);
}

void FTiltFiveHMD::ConditionalInitializeGlasses()
{
	if (!CurrentExclusiveGlasses)
	{
		if (!CurrentReservedGlasses)
		{
			CurrentReservedGlasses = RetrieveGlasses();
		}

		if (CurrentReservedGlasses)
		{
			FScopeLock ScopeLock(&ExclusiveGroup1CriticalSection);
			const FTiltFiveModule& TiltFiveModule = FTiltFiveModule::Get();
			char Identifier[T5_MAX_STRING_PARAM_LEN];
			size_t IdentifierSize = UE_ARRAY_COUNT(Identifier);

			ensure(t5GetGlassesIdentifier(CurrentReservedGlasses, Identifier, &IdentifierSize) == T5_SUCCESS);

			ET5ConnectionState ConnectionState = kT5_ConnectionState_Disconnected;

			// Try claiming the glasses for usage
			FT5Result Result = t5ReserveGlasses(CurrentReservedGlasses, TiltFiveModule.GetApplicationDisplayName());

			if (Result == T5_SUCCESS)
			{
				Result = t5GetGlassesConnectionState(CurrentReservedGlasses, &ConnectionState);
				if (Result != T5_SUCCESS)
				{
					UE_LOG(LogTiltFive, Error, TEXT("Failed to get connection state: %S"), t5GetResultMessage(Result));
				}
				UE_LOG(LogTiltFive, Log, TEXT("Glasses made exclusive: %S"), Identifier);

				// If we had some old glasses, destroy them
				if (CurrentExclusiveGlasses)
				{
					t5DestroyGlasses(&CurrentExclusiveGlasses);
					CurrentExclusiveGlasses = nullptr;
				}

				CurrentExclusiveGlasses = CurrentReservedGlasses;
				CurrentReservedGlasses = nullptr;

				int32 RetryLimit = 10;
				for (;;)
				{
					Result = t5EnsureGlassesReady(CurrentExclusiveGlasses);
					if (Result == T5_SUCCESS)
					{
						break;
					}
					else if (RetryLimit <= 0)
					{
						UE_LOG(LogTiltFive,
							Error,
							TEXT("Failed to ensure glasses are ready and reached retry limit. "
								 "Releasing glasses."));
						t5ReleaseGlasses(CurrentExclusiveGlasses);
						return;
					}
					else if (Result == T5_ERROR_TRY_AGAIN)
					{
						UE_LOG(LogTiltFive, Log, TEXT("Glasses not ready, trying again: %S"), Identifier);
						RetryLimit--;
						// Sleep for 100ms to wait for glasses to wake.
						FPlatformProcess::SleepNoStats(0.1f);
						continue;
					}
					else
					{
						UE_LOG(LogTiltFive,
							Error,
							TEXT("Failed to ensure glasses are ready. Releasing glasses.: %S"),
							t5GetResultMessage(Result));
						t5ReleaseGlasses(CurrentExclusiveGlasses);
						return;
					}
				}

				const ET5GlassesParam Parameter = kT5_ParamGlasses_Float_IPD;

				double IPDValue;

				Result = t5GetGlassesFloatParam(CurrentExclusiveGlasses, 0, Parameter, &IPDValue);

				if (Result == T5_SUCCESS)
				{
					bVersionCompatible = true;
					CachedTiltFiveIPD = static_cast<float>(IPDValue);
				}
				else if (Result == T5_ERROR_SERVICE_INCOMPATIBLE)
				{
					bVersionCompatible = false;
					UE_LOG(LogTiltFive, Error, TEXT("Version incompatible, needs service upgrade"));
				}
				else
				{
					UE_LOG(LogTiltFive, Error, TEXT("Failed to retrieve IPD Value: %S"), t5GetResultMessage(Result));
				}
			}
			else if (ConnectionState == kT5_ConnectionState_ExclusiveReservation)
			{
				UE_LOG(LogTiltFive, Verbose, TEXT("Glasses reserved: %S"), Identifier);
			}
			else
			{
				UE_LOG(LogTiltFive,
					Error,
					TEXT("Failed to make glasses exclusive for glasses %S: %S"),
					Identifier,
					t5GetResultMessage(Result));

				// Destroy them and re-try from the beginning, potentially with other glasses
				t5DestroyGlasses(&CurrentReservedGlasses);
				CurrentExclusiveGlasses = nullptr;
			}
		}
	}
	else
	{
		FScopeLock ScopeLock(&ExclusiveGroup1CriticalSection);
		ET5ConnectionState ConnectionState = kT5_ConnectionState_Disconnected;
		const FT5Result Result = t5GetGlassesConnectionState(CurrentExclusiveGlasses, &ConnectionState);
		if (Result != T5_SUCCESS || ConnectionState != kT5_ConnectionState_ExclusiveConnection)
		{
			UE_LOG(LogTiltFive, Error, TEXT("Lost connection to glasses: %S"), t5GetResultMessage(Result));

			t5DestroyGlasses(&CurrentExclusiveGlasses);
			CurrentExclusiveGlasses = nullptr;
		}
	}
}

FVector2D FTiltFiveHMD::GetTextSafeRegionBounds() const
{
	return FVector2D(0.75f, 0.75f);
}

#if UE_VERSION_OLDER_THAN(4, 26, 0)
void FTiltFiveHMD::UpdateSplashScreen()
{
}
#endif

FT5GlassesPtr FTiltFiveHMD::GetCurrentExclusiveGlasses() const
{
	return CurrentExclusiveGlasses;
}

FT5GameboardType FTiltFiveHMD::GetCurrentGameboardType() const
{
	return CurrentGameboardType;
}

#if UE_VERSION_OLDER_THAN(5, 0, 0)
bool FTiltFiveHMD::GetRelativeEyePose(int32 DeviceId, EStereoscopicPass Eye, FQuat& OutOrientation, FVector& OutPosition)
{
	OutOrientation = FQuat::Identity;
	OutPosition = FVector::ZeroVector;
	if (DeviceId == IXRTrackingSystem::HMDDeviceId && (Eye == eSSP_LEFT_EYE || Eye == eSSP_RIGHT_EYE))
	{
		OutPosition = FVector(
			0, (Eye == EStereoscopicPass::eSSP_LEFT_EYE ? -.5 : .5) * GetInterpupillaryDistance() * GetWorldToMetersScale(), 0);
		return true;
	}
	else
	{
		return false;
	}
}
#else
bool FTiltFiveHMD::GetRelativeEyePose(int32 DeviceId, int32 ViewIndex, FQuat& OutOrientation, FVector& OutPosition)
{
	OutOrientation = FQuat::Identity;
	OutPosition = FVector::ZeroVector;
	if (DeviceId == IXRTrackingSystem::HMDDeviceId &&
		(ViewIndex == EStereoscopicEye::eSSE_LEFT_EYE || ViewIndex == EStereoscopicEye::eSSE_RIGHT_EYE))
	{
		OutPosition = FVector(0,
			(ViewIndex == EStereoscopicEye::eSSE_LEFT_EYE ? -.5 : .5) * GetInterpupillaryDistance() * GetWorldToMetersScale(),
			0);
		return true;
	}
	else
	{
		return false;
	}
}
#endif

bool FTiltFiveHMD::DoesSupportPositionalTracking() const
{
	return true;
}

bool FTiltFiveHMD::DoesSupportLateUpdate() const
{
	return true;
}

bool FTiltFiveHMD::HasValidTrackingPosition()
{
	FQuat Orientation;
	FVector Position;
	return GetCurrentPose(HMDDeviceId, Orientation, Position);
}

#if UE_VERSION_NEWER_THAN(4, 20, 3)
FIntPoint FTiltFiveHMD::GetIdealDebugCanvasRenderTargetSize() const
{
	FIntPoint Result = GetIdealRenderTargetSize();
	// Only one eye
	Result.X /= 2;
	return Result;
}
#endif

FT5GlassesPtr FTiltFiveHMD::RetrieveGlasses() const
{
	FScopeLock ScopeLock(&ExclusiveGroup1CriticalSection);
	static int32 CurrentGlassesIndex = 0;
	const FTiltFiveModule& TiltFiveModule = FTiltFiveModule::Get();

	FT5GlassesPtr NewGlasses = nullptr;
	TArray<char, TFixedAllocator<GMaxNumTiltFiveGlasses * T5_MAX_STRING_PARAM_LEN>> GlassesStringBuffer;
	SIZE_T BufferSize = GMaxNumTiltFiveGlasses * T5_MAX_STRING_PARAM_LEN;

	FT5Result Result = t5ListGlasses(TiltFiveModule.GetContext(), GlassesStringBuffer.GetData(), &BufferSize);

	if (Result == T5_SUCCESS)
	{
		bVersionCompatible = true;
		// Now figure out the number of glasses of this string mess
		GlassesStringBuffer.SetNumUnsafeInternal(BufferSize);

		TArray<FString> Glasses;

		char* CurrentPosition = GlassesStringBuffer.GetData();
		bool bMoreGlasses = true;
		do
		{
			const int32 Length = FCStringAnsi::Strlen(CurrentPosition);

			if (Length > 0)
			{
				Glasses.Add(FString(CurrentPosition));
				CurrentPosition += Length + 1;
			}
			else
			{
				bMoreGlasses = false;
			}
		} while (bMoreGlasses);

		if (Glasses.Num())
		{
			Result = t5CreateGlasses(
				TiltFiveModule.GetContext(), TCHAR_TO_UTF8(*Glasses[CurrentGlassesIndex % Glasses.Num()]), &NewGlasses);

			if (Result == T5_SUCCESS)
			{
				bVersionCompatible = true;
				UE_LOG(LogTiltFive, Log, TEXT("Using glasses with ID: %s"), *Glasses[CurrentGlassesIndex % Glasses.Num()]);
			}
			else if (Result == T5_ERROR_SERVICE_INCOMPATIBLE)
			{
				bVersionCompatible = false;
				UE_LOG(LogTiltFive, Error, TEXT("Version incompatible, needs service upgrade"));
			}
		}
		else
		{
			UE_LOG(LogTiltFive, VeryVerbose, TEXT("No glasses connected..."));
		}
	}
	else if (Result == T5_ERROR_SERVICE_INCOMPATIBLE)
	{
		bVersionCompatible = false;
		UE_LOG(LogTiltFive, Error, TEXT("Version incompatible, needs service upgrade"));
	}
	else
	{
		// T5_ERROR_NO_SERVICE or T5_ERROR_IO_FAILURE apparently means it is starting up internally,
		// and we get it soon(tm)
		UE_CLOG(Result != T5_ERROR_NO_SERVICE && Result != T5_ERROR_IO_FAILURE,
			LogTiltFive,
			Error,
			TEXT("Failed to retrieve glasses: %S"),
			t5GetResultMessage(Result));
	}

	// Next time, try the next index
	++CurrentGlassesIndex;

	return NewGlasses;
}

class FXRRenderBridge* FTiltFiveHMD::GetActiveRenderBridge_GameThread(bool bUseSeparateRenderTarget)
{
	return CustomPresent;
}

FName FTiltFiveHMD::GetSystemName() const
{
	return TiltFiveSystemName;
}

bool FTiltFiveHMD::IsHMDConnected()
{
	if (CurrentExclusiveGlasses)
	{
		return true;
	}
	const FTiltFiveModule& TiltFiveModule = FTiltFiveModule::Get();
	FT5GlassesPtr NewGlasses = nullptr;
	TArray<char, TFixedAllocator<GMaxNumTiltFiveGlasses * T5_MAX_STRING_PARAM_LEN>> GlassesStringBuffer;
	SIZE_T BufferSize = GMaxNumTiltFiveGlasses * T5_MAX_STRING_PARAM_LEN;

	{
		FScopeLock ScopeLock(&ExclusiveGroup1CriticalSection);
		const FT5Result Result = t5ListGlasses(TiltFiveModule.GetContext(), GlassesStringBuffer.GetData(), &BufferSize);

		if (Result == T5_ERROR_SERVICE_INCOMPATIBLE)
		{
			bVersionCompatible = false;
			UE_LOG(LogTiltFive, Error, TEXT("Version incompatible, needs service upgrade"));
		}
		bVersionCompatible = true;
		return Result == T5_SUCCESS && BufferSize > 0;
	}
}

bool FTiltFiveHMD::IsHMDEnabled() const
{
	return CurrentExclusiveGlasses != nullptr;
}

bool FTiltFiveHMD::IsVersionCompatible() const
{
	return bVersionCompatible;
}

void FTiltFiveHMD::EnableHMD(bool bEnable /*= true*/)
{
	unimplemented();
}

bool FTiltFiveHMD::GetHMDMonitorInfo(MonitorInfo& OutMonitorInfo)
{
	OutMonitorInfo.MonitorName = FString("Tilt Five Window");
	OutMonitorInfo.MonitorId = 0;
	OutMonitorInfo.DesktopX = 0;
	OutMonitorInfo.DesktopY = 0;

	const FIntPoint RenderTargetSize = GetIdealRenderTargetSize();
	OutMonitorInfo.ResolutionX = RenderTargetSize.X;
	OutMonitorInfo.ResolutionY = RenderTargetSize.Y;
	OutMonitorInfo.WindowSizeX = RenderTargetSize.X;
	OutMonitorInfo.WindowSizeY = RenderTargetSize.Y;

	return true;
}

void FTiltFiveHMD::GetFieldOfView(float& InOutHFOVInDegrees, float& InOutVFOVInDegrees) const
{
	InOutHFOVInDegrees = FOV;
	InOutVFOVInDegrees = FOV;
}

void FTiltFiveHMD::SetInterpupillaryDistance(float NewInterpupillaryDistance)
{
	// Most XR plugins appear to ignore requests to set the IPD. Not supporting having this override the IPD that the user has set
	// in the Tilt Five Control Panel simplifies making the CachedTiltFiveIPD safe-ish to access from different threads. We should
	// really be snapping it once per frame on the Game thread and then plumbing it through to the Render thread each frame, but
	// stuffing it in an atomic should be good enough for now.
}

float FTiltFiveHMD::GetInterpupillaryDistance() const
{
	// The interpupillary distance this is supposed to return appears to be in physical meters rather than Unreal units. It's not
	// entirely clear that's the case, but SetInterpupillaryDistance is documented as "Accessors to modify the interpupillary
	// distance (meters)" and many plugins default to or are hard-coded to return 0.064.
	return CachedTiltFiveIPD;
}

bool FTiltFiveHMD::IsChromaAbCorrectionEnabled() const
{
	return false;
}

bool FTiltFiveHMD::IsStereoEnabled() const
{
	return bEnableStereo;
}

bool FTiltFiveHMD::EnableStereo(bool stereo /*= true*/)
{
	bEnableStereo = stereo;
	return true;
}

#if UE_VERSION_OLDER_THAN(5, 0, 0)
void FTiltFiveHMD::AdjustViewRect(enum EStereoscopicPass StereoPass, int32& X, int32& Y, uint32& SizeX, uint32& SizeY) const
{
	if (StereoPass == eSSP_FULL)
	{
		return;
	}
	const int32 EyeIndex = StereoPass == EStereoscopicPass::eSSP_RIGHT_EYE ? 1 : 0;
	FIntRect Rect = EyeInfos[EyeIndex].SourceEyeRect;
	X = Rect.Min.X;
	Y = Rect.Min.Y;
	SizeX = Rect.Width();
	SizeY = Rect.Height();
}
#else
void FTiltFiveHMD::AdjustViewRect(int32 ViewIndex, int32& X, int32& Y, uint32& SizeX, uint32& SizeY) const
{
	if (ViewIndex != EStereoscopicEye::eSSE_RIGHT_EYE && ViewIndex != EStereoscopicEye::eSSE_LEFT_EYE)
	{
		return;
	}
	const int32 EyeIndex = ViewIndex == EStereoscopicEye::eSSE_RIGHT_EYE ? 1 : 0;
	FIntRect Rect = EyeInfos[EyeIndex].SourceEyeRect;
	X = Rect.Min.X;
	Y = Rect.Min.Y;
	SizeX = Rect.Width();
	SizeY = Rect.Height();
}
#endif

#if UE_VERSION_OLDER_THAN(5, 0, 0)
FMatrix FTiltFiveHMD::GetStereoProjectionMatrix(const enum EStereoscopicPass StereoPassType) const
{
	const int32 EyeIndex = StereoPassType == EStereoscopicPass::eSSP_RIGHT_EYE ? 1 : 0;

	const float HalfFov = FMath::DegreesToRadians(FOV) / 2.f;
	const float HalfFovTan = FMath::Tan(HalfFov);
	const float InWidth = EyeInfos[EyeIndex].DestEyeRect.Width();
	const float InHeight = EyeInfos[EyeIndex].DestEyeRect.Height();
	const float XS = 1.0f / HalfFovTan;
	const float YS = InWidth / HalfFovTan / InHeight;

	const float InNearZ = GNearClippingPlane;
	return FMatrix(FPlane(XS, 0.0f, 0.0f, 0.0f),
		FPlane(0.0f, YS, 0.0f, 0.0f),
		FPlane(0.0f, 0.0f, 0.0f, 1.0f),
		FPlane(0.0f, 0.0f, InNearZ, 0.0f));
}
#else
FMatrix FTiltFiveHMD::GetStereoProjectionMatrix(const int32 ViewIndex) const
{
	const int32 EyeIndex = ViewIndex == EStereoscopicEye::eSSE_RIGHT_EYE ? 1 : 0;

	const float HalfFov = FMath::DegreesToRadians(FOV) / 2.f;
	const float HalfFovTan = FMath::Tan(HalfFov);
	const float InWidth = EyeInfos[EyeIndex].DestEyeRect.Width();
	const float InHeight = EyeInfos[EyeIndex].DestEyeRect.Height();
	const float XS = 1.0f / HalfFovTan;
	const float YS = InWidth / HalfFovTan / InHeight;

	const float InNearZ = GNearClippingPlane;
	return FMatrix(FPlane(XS, 0.0f, 0.0f, 0.0f),
		FPlane(0.0f, YS, 0.0f, 0.0f),
		FPlane(0.0f, 0.0f, 0.0f, 1.0f),
		FPlane(0.0f, 0.0f, InNearZ, 0.0f));
}
#endif

void FTiltFiveHMD::InitCanvasFromView(class FSceneView* InView, class UCanvas* Canvas)
{
	FSceneView HMDView(*InView);

	Canvas->UpdateSafeZoneData();

	const FQuat DeltaRot = HMDView.BaseHmdOrientation.Inverse() * Canvas->HmdOrientation;
	HMDView.ViewRotation = FRotator(HMDView.ViewRotation.Quaternion() * DeltaRot);
	HMDView.UpdateViewMatrix();
	Canvas->ViewProjectionMatrix = HMDView.ViewMatrices.GetViewProjectionMatrix();
}

class IHeadMountedDisplay* FTiltFiveHMD::GetHMDDevice()
{
	return this;
}

class TSharedPtr<class IStereoRendering, ESPMode::ThreadSafe> FTiltFiveHMD::GetStereoRenderingDevice()
{
	return SharedThis(this);
}

bool FTiltFiveHMD::ShouldUseSeparateRenderTarget() const
{
	return IsStereoEnabled();
}

uint32 FTiltFiveHMD::CreateLayer(const FLayerDesc& InLayerDesc)
{
	unimplemented();
	return INDEX_NONE;
}

void FTiltFiveHMD::DestroyLayer(uint32 LayerId)
{
	unimplemented();
}

void FTiltFiveHMD::SetLayerDesc(uint32 LayerId, const FLayerDesc& InLayerDesc)
{
	unimplemented();
}

bool FTiltFiveHMD::GetLayerDesc(uint32 LayerId, FLayerDesc& OutLayerDesc)
{
	unimplemented();
	return false;
}

void FTiltFiveHMD::MarkTextureForUpdate(uint32 LayerId)
{
	unimplemented();
}

#if UE_VERSION_NEWER_THAN(4, 21, 2)
bool FTiltFiveHMD::ShouldCopyDebugLayersToSpectatorScreen() const
{
	return false;
}
#endif

void FTiltFiveHMD::SetupViewFamily(FSceneViewFamily& InViewFamily)
{
	check(IsInGameThread());
}

void FTiltFiveHMD::SetupView(FSceneViewFamily& InViewFamily, FSceneView& InView)
{
	check(IsInGameThread());
}

void FTiltFiveHMD::BeginRenderViewFamily(FSceneViewFamily& InViewFamily)
{
	check(IsInGameThread());
	if (SpectatorScreenController != nullptr)
	{
		SpectatorScreenController->BeginRenderViewFamily();
	}
}

void FTiltFiveHMD::OnBeginRendering_RenderThread(FRHICommandListImmediate& RHICmdList, FSceneViewFamily& InViewFamily)
{
	check(IsInRenderingThread());

	UpdateCachedGlassesPose_RenderThread();
}

void FTiltFiveHMD::OnLateUpdateApplied_RenderThread(
#if ENGINE_MINOR_VERSION >= 27 || ENGINE_MAJOR_VERSION == 5
	FRHICommandListImmediate& RHICmdList,
#endif	  // ENGINE_MINOR_VERSION >= 27
	const FTransform& NewRelativeTransform)
{
	check(IsInRenderingThread());

	MaybeRelativeGlassesTransform_RenderThread = NewRelativeTransform;
}

void FTiltFiveHMD::PreRenderViewFamily_RenderThread(
#if UE_VERSION_OLDER_THAN(5, 1, 0)
	FRHICommandListImmediate& RHICmdList,
#else
	FRDGBuilder& GraphBuilder,
#endif
	FSceneViewFamily& InViewFamily)
{
	check(IsInRenderingThread());

	if (!InViewFamily.RenderTarget->GetRenderTargetTexture())
	{
		return;
	}

	if (SpectatorScreenController)
	{
		SpectatorScreenController->UpdateSpectatorScreenMode_RenderThread();
	}
}

void FTiltFiveHMD::PreRenderView_RenderThread(
#if UE_VERSION_OLDER_THAN(5, 1, 0)
	FRHICommandListImmediate& RHICmdList,
#else
	FRDGBuilder& GraphBuilder,
#endif
	FSceneView& InView)
{
	check(IsInRenderingThread());
}

int32 FTiltFiveHMD::GetPriority() const
{
	// We want to run before the FDefaultXRCamera's view extension, as well as the SteamVR and Oculus VR headsets
	return 1;
}

FTiltFiveCustomPresent::FTiltFiveCustomPresent(FTiltFiveHMD& InTiltFiveHMD)
	: HMD(InTiltFiveHMD)
{
}

bool FTiltFiveCustomPresent::NeedsNativePresent()
{
	return true;
}

bool FTiltFiveCustomPresent::Present(int32& InOutSyncInterval)
{
	// XXX: This currently passes, but is this always going to be called on the render thread?
	check(IsInRenderingThread());
	// XXX: Important note: Custom Present currently runs on the "RHI" thread, which, when using
	// DirectX 11 is synonymous with the render thread. Unreal 5.0 forward supported DirectX 12
	// by default. When this happens, we are no longer on the render thread, and are wholly on
	// the RHI thread. We don't currently support DirectX 12, but once we do, we need to refactor
	// this section to make sure the pose is accessible on the RHI thread specifically, separate
	// from the render thread.

	if (!HMD.MaybeRelativeGlassesTransform_RenderThread.IsSet())
	{
		return false;
	}

	FQuat GlassesOrientation = HMD.MaybeRelativeGlassesTransform_RenderThread->GetRotation();
	FVector GlassesPosition = HMD.MaybeRelativeGlassesTransform_RenderThread->GetTranslation();

	FT5FrameInfo FrameInfo;

	const float WorldToMetersScale = HMD.GetWorldToMetersScale();

	float IPD_UWRLD = HMD.GetInterpupillaryDistance() * WorldToMetersScale;

	FTiltFiveHMD::ConvertPositionToHardware(
		GlassesPosition - GlassesOrientation.GetRightVector() * IPD_UWRLD * 0.5f, FrameInfo.posLVC_GBD, WorldToMetersScale);
	FTiltFiveHMD::ConvertPositionToHardware(
		GlassesPosition + GlassesOrientation.GetRightVector() * IPD_UWRLD * 0.5f, FrameInfo.posRVC_GBD, WorldToMetersScale);

	FTiltFiveHMD::ConvertRotationToHardware(GlassesOrientation, FrameInfo.rotToLVC_GBD);
	FTiltFiveHMD::ConvertRotationToHardware(GlassesOrientation, FrameInfo.rotToRVC_GBD);

	FrameInfo.vci.startX_VCI = -FMath::Tan(HMD.FOV * (0.5f * PI / 180.0f));
	FrameInfo.vci.startY_VCI = FrameInfo.vci.startX_VCI / HMD.WidthToHeight;
	FrameInfo.vci.width_VCI = -2.0f * FrameInfo.vci.startX_VCI;
	FrameInfo.vci.height_VCI = -2.0f * FrameInfo.vci.startY_VCI;

	FrameInfo.isUpsideDown = true;
	FrameInfo.isSrgb = EnumHasAnyFlags(HMD.EyeInfos[0].BufferedSRVRHI->GetFlags(), TexCreate_SRGB) != 0;
	FrameInfo.leftTexHandle = HMD.EyeInfos[0].BufferedSRVRHI->GetNativeResource();
	FrameInfo.rightTexHandle = HMD.EyeInfos[1].BufferedSRVRHI->GetNativeResource();
	FrameInfo.texWidth_PIX = HMD.EyeInfos[0].BufferedSRVRHI->GetSizeX();
	FrameInfo.texHeight_PIX = HMD.EyeInfos[0].BufferedSRVRHI->GetSizeY();

	if (HMD.CurrentExclusiveGlasses)
	{
		FScopeLock ScopeLock(&HMD.ExclusiveGroup1CriticalSection);

		if (HMD.CurrentExclusiveGlasses != HMD.GraphicsInitializedGlasses)
		{
			// Obtain a graphics context
			ET5GraphicsAPI GraphicsAPI{};
			void* GraphicsContext{};
			const FString RHIName = GDynamicRHI->GetName();

			GraphicsAPI = kT5_GraphicsApi_None;
			GraphicsContext = nullptr;
			if (RHIName == TEXT("D3D11"))
			{
				GraphicsAPI = kT5_GraphicsApi_D3D11;
				GraphicsContext = GDynamicRHI->RHIGetNativeDevice();
			}
			else if (RHIName == TEXT("OpenGL"))
			{
				GraphicsAPI = kT5_GraphicsApi_GL;
			}

			FT5Result Result = t5InitGlassesGraphicsContext(HMD.CurrentExclusiveGlasses, GraphicsAPI, GraphicsContext);
			if (Result != T5_SUCCESS)
			{
				UE_LOG(LogTiltFive, Error, TEXT("Failed to initialize graphics context"));
				return false;
			}

			HMD.GraphicsInitializedGlasses = HMD.CurrentExclusiveGlasses;
		}

		const FT5Result Result = t5SendFrameToGlasses(HMD.CurrentExclusiveGlasses, &FrameInfo);

		UE_CLOG(Result != T5_SUCCESS, LogTiltFive, Error, TEXT("Failed to send frame: %S"), t5GetResultMessage(Result));
	}
	return true;
}

void FTiltFiveCustomPresent::UpdateViewport(const class FViewport& Viewport, class FRHIViewport* InViewportRHI)
{
	FXRRenderBridge::UpdateViewport(Viewport, InViewportRHI);

	if (HMD.EyeInfos[0].BufferedRTRHI)
	{
		return;
	}
	const FIntPoint Size = HMD.GetIdealRenderTargetSize();

	FRHIResourceCreateInfo CreateInfo{TEXT("TiltFiveCreateInfo")};
	const int32 EyeSizeX = Size.X / 2;
	const int32 EyeSizeY = Size.Y;
	HMD.WidthToHeight = (float) EyeSizeX / (float) EyeSizeY;

	for (int32 EyeIndex = 0; EyeIndex < NumEyeRenderTargets; ++EyeIndex)
	{
		FTiltFiveEyeInfo& EyeInfo = HMD.EyeInfos[EyeIndex];
#if !UE_BUILD_SHIPPING
		EyeInfo.DebugName = FString::Printf(TEXT("TiltFiveEyeRT%d"), EyeIndex);
		CreateInfo.DebugName = *EyeInfo.DebugName;
#endif
		EyeInfo.SourceEyeRect = FIntRect(EyeSizeX * EyeIndex, 0, EyeSizeX * EyeIndex + EyeSizeX, EyeSizeY);
		EyeInfo.DestEyeRect = FIntRect(0, 0, EyeSizeX, EyeSizeY);
		RHICreateTargetableShaderResource2D(EyeSizeX,
			EyeSizeY,
			PF_R8G8B8A8,
			1,
			TexCreate_None,
			TexCreate_RenderTargetable,
			false,
			CreateInfo,
			EyeInfo.BufferedRTRHI,
			EyeInfo.BufferedSRVRHI);
	}
}
