// Copyright 2022 Tilt Five, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#pragma once

#include "CoreMinimal.h"
#include "HeadMountedDisplayBase.h"
#include "IStereoLayers.h"
#include "Misc/EngineVersionComparison.h"
#include "Runtime/Launch/Resources/Version.h"
#include "SceneViewExtension.h"
#include "TiltFive.h"
#include "XRRenderBridge.h"
#include "XRRenderTargetManager.h"

#include <atomic>

namespace ETiltFiveDeviceId
{
enum Type
{
	Invalid = INDEX_NONE,
	// 0 should represent the default HMD Device, see IXRTrackingSystem::HMDDeviceId
	Glasses = 0,
	Glasses_0 = 0,
};
}

class FTiltFiveHMD;

class FTiltFiveCustomPresent : public FXRRenderBridge
{
public:
	FTiltFiveCustomPresent(FTiltFiveHMD& InTiltFiveHMD);
	virtual bool NeedsNativePresent() override;
	virtual bool Present(int32& InOutSyncInterval) override;

	void UpdateViewport(const class FViewport& Viewport, class FRHIViewport* InViewportRHI) override;

private:
	FTiltFiveHMD& HMD;
};

constexpr int32 NumEyeRenderTargets = 2;

struct FTiltFiveEyeInfo
{
	FTexture2DRHIRef BufferedRTRHI;
	FTexture2DRHIRef BufferedSRVRHI;
	FIntRect SourceEyeRect;
	FIntRect DestEyeRect;
#if !UE_BUILD_SHIPPING
	FString DebugName;
#endif
};

struct FTiltFiveWorldState
{
	float WorldToMetersScale;
};

typedef TSharedPtr<FTiltFiveWorldState, ESPMode::ThreadSafe> FTiltFiveWorldStatePtr;

/**
 * Tilt Five Head Mounted Display Interface Implementation
 */
class TILTFIVE_API FTiltFiveHMD : public FHeadMountedDisplayBase,
								  public FXRRenderTargetManager,
								  public IStereoLayers,
								  public FSceneViewExtensionBase
{
public:
	FTiltFiveHMD(const FAutoRegister&);
	virtual ~FTiltFiveHMD() override;

	void Startup();
	void ConditionalInitializeGlasses();

	static FQuat ConvertFromHardware(const FQuat& Rotation);
	static FQuat ConvertToHardware(const FQuat& Rotation);
	static FVector ConvertFromHardware(const FVector& Rotation, float WorldToMetersScale);
	static FVector ConvertToHardware(const FVector& Rotation, float WorldToMetersScale);

	// Someone thought anonymous structs were a good idea.
	template <typename T> static FQuat ConvertRotationFromHardware(const T& Rotation)
	{
		const FQuat HardwareRotation(Rotation.x, Rotation.y, Rotation.z, Rotation.w);
		return ConvertFromHardware(HardwareRotation);
	}

	template <typename T> static FVector ConvertPositionFromHardware(const T& Position, float WorldToMetersScale)
	{
		const FVector HardwarePosition(Position.x, Position.y, Position.z);
		return ConvertFromHardware(HardwarePosition, WorldToMetersScale);
	}

	template <typename T> static void ConvertRotationToHardware(const FQuat& SourceRotation, T& OutHardwareRotation)
	{
		const FQuat HardwareRotation(ConvertToHardware(SourceRotation));
		OutHardwareRotation.x = HardwareRotation.X;
		OutHardwareRotation.y = HardwareRotation.Y;
		OutHardwareRotation.z = HardwareRotation.Z;
		OutHardwareRotation.w = HardwareRotation.W;
	}

	template <typename T>
	static void ConvertPositionToHardware(const FVector& SourcePosition, T& OutHardwarePosition, float WorldToMetersScale)
	{
		const FVector HardwarePosition(ConvertToHardware(SourcePosition, WorldToMetersScale));
		OutHardwarePosition.x = HardwarePosition.X;
		OutHardwarePosition.y = HardwarePosition.Y;
		OutHardwarePosition.z = HardwarePosition.Z;
	}

	// FHeadMountedDisplayBase Interface
#if UE_VERSION_NEWER_THAN(4, 25, 4)
	virtual int32 GetXRSystemFlags() const override;
	virtual bool AllocateRenderTargetTexture(uint32 Index,
		uint32 SizeX,
		uint32 SizeY,
		uint8 Format,
		uint32 NumMips,
		ETextureCreateFlags Flags,
		ETextureCreateFlags TargetableTextureFlags,
		FTexture2DRHIRef& OutTargetableTexture,
		FTexture2DRHIRef& OutShaderResourceTexture,
		uint32 NumSamples = 1) override;
#endif
#if UE_VERSION_NEWER_THAN(4, 22, 3)
	virtual EHMDTrackingOrigin::Type GetTrackingOrigin() const override;
#endif
#if UE_VERSION_NEWER_THAN(4, 21, 2)
	virtual bool ShouldCopyDebugLayersToSpectatorScreen() const override;
#endif

#if UE_VERSION_NEWER_THAN(4, 20, 3)
	virtual FIntPoint GetIdealDebugCanvasRenderTargetSize() const override;
#endif
#if UE_VERSION_NEWER_THAN(4, 20, 3)
	virtual void CopyTexture_RenderThread(FRHICommandListImmediate& RHICmdList,
		FRHITexture2D* SrcTexture,
		FIntRect SrcRect,
		FRHITexture2D* DstTexture,
		FIntRect DstRect,
		bool bClearBlack,
		bool bNoAlpha) const override;
#else
	virtual void CopyTexture_RenderThread(FRHICommandListImmediate& RHICmdList,
		FRHITexture2D* SrcTexture,
		FIntRect SrcRect,
		FRHITexture2D* DstTexture,
		FIntRect DstRect,
		bool bClearBlack) const override;
#endif

	virtual bool EnumerateTrackedDevices(TArray<int32>& OutDevices, EXRTrackedDeviceType Type = EXRTrackedDeviceType::Any) override;
	virtual bool GetCurrentPose(int32 DeviceId, FQuat& OutOrientation, FVector& OutPosition) override;
	virtual float GetWorldToMetersScale() const override;
	virtual void ResetOrientationAndPosition(float Yaw = 0.f) override;
	virtual FName GetSystemName() const override;
	virtual bool IsHMDConnected() override;
	virtual bool IsHMDEnabled() const override;
	virtual bool IsVersionCompatible() const;
	virtual void EnableHMD(bool bEnable = true) override;
	virtual bool GetHMDMonitorInfo(MonitorInfo&) override;
	virtual void GetFieldOfView(float& InOutHFOVInDegrees, float& InOutVFOVInDegrees) const override;
	virtual void SetInterpupillaryDistance(float NewInterpupillaryDistance) override;
	virtual float GetInterpupillaryDistance() const override;
	virtual bool IsChromaAbCorrectionEnabled() const override;
	virtual bool IsStereoEnabled() const override;
	virtual bool EnableStereo(bool stereo = true) override;
	virtual void InitCanvasFromView(class FSceneView* InView, class UCanvas* Canvas) override;
	virtual FIntPoint GetIdealRenderTargetSize() const override;

	virtual class IHeadMountedDisplay* GetHMDDevice() override;
	virtual class TSharedPtr<class IStereoRendering, ESPMode::ThreadSafe> GetStereoRenderingDevice() override;
	// /FHeadMountedDisplayBase interface

	virtual bool ShouldUseSeparateRenderTarget() const override;
	virtual uint32 CreateLayer(const FLayerDesc& InLayerDesc) override;
	virtual void DestroyLayer(uint32 LayerId) override;
	virtual void SetLayerDesc(uint32 LayerId, const FLayerDesc& InLayerDesc) override;
	virtual bool GetLayerDesc(uint32 LayerId, FLayerDesc& OutLayerDesc) override;
	virtual void MarkTextureForUpdate(uint32 LayerId) override;
	virtual void SetupViewFamily(FSceneViewFamily& InViewFamily) override;
	virtual void SetupView(FSceneViewFamily& InViewFamily, FSceneView& InView) override;
	virtual void BeginRenderViewFamily(FSceneViewFamily& InViewFamily) override;
	virtual void PreRenderViewFamily_RenderThread(
#if UE_VERSION_OLDER_THAN(5, 1, 0)
		FRHICommandListImmediate& RHICmdList,
#else
		FRDGBuilder& GraphBuilder,
#endif
		FSceneViewFamily& InViewFamily) override;
	virtual void PreRenderView_RenderThread(
#if UE_VERSION_OLDER_THAN(5, 1, 0)
		FRHICommandListImmediate& RHICmdList,
#else
		FRDGBuilder& GraphBuilder,
#endif
		FSceneView& InView) override;
	virtual int32 GetPriority() const override;

#if UE_VERSION_OLDER_THAN(5, 1, 0)
	virtual bool IsActiveThisFrame(class FViewport* InViewport) const override;
#endif
	virtual void RenderTexture_RenderThread(class FRHICommandListImmediate& RHICmdList,
#if UE_VERSION_OLDER_THAN(5, 1, 0)
		class FRHITexture2D* BackBuffer,
		class FRHITexture2D* SrcTexture,
#else
		class FRHITexture* BackBuffer,
		class FRHITexture* SrcTexture,
#endif
		FVector2D WindowSize) const override;

	virtual IStereoRenderTargetManager* GetRenderTargetManager() override;
	virtual bool GetHMDDistortionEnabled(EShadingPath ShadingPath) const override;
	virtual void SetTrackingOrigin(EHMDTrackingOrigin::Type NewOrigin) override;
	virtual bool OnStartGameFrame(FWorldContext& WorldContext) override;
	virtual FVector2D GetTextSafeRegionBounds() const override;
	virtual void OnBeginRendering_RenderThread(FRHICommandListImmediate& RHICmdList, FSceneViewFamily& InViewFamily) override;
	virtual void OnLateUpdateApplied_RenderThread(
#if ENGINE_MINOR_VERSION >= 27 || ENGINE_MAJOR_VERSION == 5
		FRHICommandListImmediate& RHICmdList,
#endif	  // ENGINE_MINOR_VERSION >= 27
		const FTransform& NewRelativeTransform) override;

	static const FName TiltFiveSystemName;
	// Critical section for the exclusive groups inside the tilt five API
	mutable FCriticalSection ExclusiveGroup1CriticalSection;
	mutable FCriticalSection ExclusiveGroup2CriticalSection;

#if UE_VERSION_OLDER_THAN(4, 26, 0)
	virtual void UpdateSplashScreen() override;
#endif

	FT5GlassesPtr GetCurrentExclusiveGlasses() const;
	FT5GameboardType GetCurrentGameboardType() const;

#if UE_VERSION_OLDER_THAN(5, 0, 0)
	virtual bool GetRelativeEyePose(int32 DeviceId, EStereoscopicPass Eye, FQuat& OutOrientation, FVector& OutPosition) override;
	virtual void AdjustViewRect(enum EStereoscopicPass StereoPass, int32& X, int32& Y, uint32& SizeX, uint32& SizeY) const override;
	virtual FMatrix GetStereoProjectionMatrix(const enum EStereoscopicPass StereoPassType) const override;
#else
	virtual bool GetRelativeEyePose(int32 DeviceId, int32 ViewIndex, FQuat& OutOrientation, FVector& OutPosition) override;
	virtual void AdjustViewRect(int32 ViewIndex, int32& X, int32& Y, uint32& SizeX, uint32& SizeY) const override;
	virtual FMatrix GetStereoProjectionMatrix(const int32 ViewIndex) const override;
#endif
	virtual bool DoesSupportPositionalTracking() const override;
	virtual bool DoesSupportLateUpdate() const override;
	virtual bool HasValidTrackingPosition() override;

	void UpdateCachedGlassesPose_RenderThread();
	void UpdateCachedGlassesPose_GameThread();

private:
	bool bEnableStereo = true;
	mutable bool bVersionCompatible = false;

	class IRendererModule* RendererModule;

	FTiltFiveEyeInfo EyeInfos[2];
	float FOV = 70.0f;
	float WidthToHeight = 1.0f;

	TRefCountPtr<FTiltFiveCustomPresent> CustomPresent;

	FT5GlassesPtr CurrentReservedGlasses = nullptr;
	FT5GlassesPtr CurrentExclusiveGlasses = nullptr;
	FT5GlassesPtr GraphicsInitializedGlasses = nullptr;

	FT5GameboardType CurrentGameboardType = FT5GameboardType::kT5_GameboardType_None;

	float LastGlassesUpdate = -1.0f;
	std::atomic<float> CachedTiltFiveIPD = 0.064f;

	FT5GlassesPtr RetrieveGlasses() const;
	FTiltFiveWorldStatePtr CurrentWorldState;
	FTiltFiveWorldStatePtr RenderThreadWorldState;

	// Render thread cached glasses pose.
	bool CachedGlassesPoseIsValid_RenderThread = false;
	FQuat CachedGlassesOrientation_RenderThread;
	FVector CachedGlassesPosition_RenderThread;
	FT5GameboardType CachedGameboardType_RenderThread;

	// Game thread cached glasses pose.
	bool CachedGlassesPoseIsValid_GameThread = false;
	FQuat CachedGlassesOrientation_GameThread;
	FVector CachedGlassesPosition_GameThread;
	FT5GameboardType CachedGameboardType_GameThread;

	// The UWRLD pose of the glasses relative to its parent (i.e. the AR root) that is being used to render the current frame. This
	// gets set to an 'early' pose by OnStartGameFrame and may get updated by OnLateUpdateApplied_RenderThread if a late update is
	// done by the DefaultXRCamera. If no glasses pose was available at the time this gets set then it is in the "unset" state.
	TOptional<FTransform> MaybeRelativeGlassesTransform_RenderThread;

protected:
	virtual class FXRRenderBridge* GetActiveRenderBridge_GameThread(bool bUseSeparateRenderTarget) override;

	friend class FTiltFiveCustomPresent;
};
